# Copyright 2016-2017 Nitor Creations Oy
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Main module for nameless-deploy-tools
"""
import base64

VERSION = "1.298"

PATH_COMMANDS = [
    "bin/create-shell-archive.sh",
    "bin/ensure-letsencrypt-certs.sh",
    "bin/lastpass-fetch-notes.sh",
    "bin/lpssh",
    "bin/encrypt-and-mount.sh",
    "bin/mount-and-format.sh",
    "bin/setup-fetch-secrets.sh",
    "bin/ssh-hostkeys-collect.sh",
]
NDT_AND_CONSOLE = [
    "n-include=n_utils.cli:resolve_include",
    "n-include-all=n_utils.cli:resolve_all_includes",
    "cf-logs-to-cloudwatch=ec2_utils.cli:log_to_cloudwatch",
    "logs-to-cloudwatch=ec2_utils.cli:log_to_cloudwatch",
    "associate-eip=n_utils.cli:associate_eip",
    "signal-cf-status=ec2_utils.cli:cf_signal_status",
    "ec2-associate-eip=n_utils.cli:associate_eip",
]
NDT_ONLY = [
    "account-id=ec2_utils.cli:account_id",
    "add-deployer-server=n_utils.cli:add_deployer_server",
    "assume-role=n_utils.cli:assume_role",
    "assumed-role-name=n_utils.cli:cli_assumed_role_name",
    "azure-ensure-group=n_utils.cli:azure_ensure_group",
    "azure-ensure-management-group=n_utils.cli:azure_ensure_management_group",
    "azure-location=n_utils.cli:azure_location",
    "azure-template-parameters=n_utils.cli:azure_template_parameters",
    "bw-store-aws-cli-creds=n_utils.profile_util:cli_store_bw_profile",
    "cf-delete-stack=n_utils.cli:delete_stack",
    "cf-follow-logs=n_utils.cli:tail_stack_logs",
    "cf-get-parameter=ec2_utils.cli:cf_get_parameter",
    "cf-logical-id=ec2_utils.cli:cf_logical_id",
    "cf-region=ec2_utils.cli:cf_region",
    "cf-signal-status=ec2_utils.cli:cf_signal_status",
    "cf-stack-name=ec2_utils.cli:cf_stack_name",
    "cf-stack-id=ec2_utils.cli:cf_stack_id",
    "create-account=n_utils.cli:cli_create_account",
    "create-stack=n_utils.cf_bootstrap:create_stack",
    "detach-volume=ec2_utils.cli:detach_volume",
    "ec2-clean-snapshots=n_utils.cli:clean_snapshots",
    "ec2-get-tag=ec2_utils.cli:get_tag",
    "ec2-get-userdata=ec2_utils.cli:get_userdata",
    "ec2-instance-id=ec2_utils.cli:instance_id",
    "ec2-region=ec2_utils.cli:region",
    "ec2-wait-for-metadata=ec2_utils.cli:wait_for_metadata",
    "ecr-ensure-repo=n_utils.cli:cli_ecr_ensure_repo",
    "ecr-repo-uri=n_utils.cli:cli_ecr_repo_uri",
    "enable-profile=n_utils.profile_util:cli_enable_profile",
    "get-images=n_utils.cli:cli_get_images",
    "interpolate-file=n_utils.cli:cli_interpolate_file",
    "json-to-yaml=n_utils.cli:json_to_yaml",
    "latest-snapshot=ec2_utils.cli:latest_snapshot",
    "list-components=n_utils.cli:cli_list_components",
    "list-file-to-json=n_utils.cli:list_file_to_json",
    "list-jobs=n_utils.cli:cli_list_jobs",
    "load-parameters=n_utils.cli:cli_load_parameters",
    "logs=ec2_utils.cli:get_logs",
    "mfa-add-token=n_utils.cli:cli_mfa_add_token",
    "mfa-backup=n_utils.cli:cli_mfa_backup_tokens",
    "mfa-code=n_utils.cli:cli_mfa_code",
    "mfa-delete-token=n_utils.cli:cli_mfa_delete_token",
    "mfa-qrcode=n_utils.cli:cli_mfa_to_qrcode",
    "print-aws-profiles=n_utils.profile_util:print_aws_profiles",
    "profile-expiry-to-env=n_utils.profile_util:profile_expiry_to_env",
    "profile-to-env=n_utils.profile_util:profile_to_env",
    "promote-image=n_utils.cli:cli_promote_image",
    "pytail=ec2_utils.cli:read_and_follow",
    "read-profile-expiry=n_utils.profile_util:cli_read_profile_expiry",
    "region=ec2_utils.cli:region",
    "register-private-dns=ec2_utils.cli:register_private_dns",
    "session-to-env=n_utils.cli:session_to_env",
    "setup-cli=n_utils.cli:setup_cli",
    "share-to-another-region=n_utils.cli:cli_share_to_another_region",
    "show-stack-params-and-outputs=n_utils.cli:show_stack_params_and_outputs",
    "show-terraform-params=n_utils.cli:show_terraform_params",
    "show-azure-params=n_utils.cli:show_azure_params",
    "snapshot-from-volume=ec2_utils.cli:snapshot_from_volume",
    "upsert-cloudfront-records=n_utils.cli:cli_upsert_cloudfront_records",
    "volume-from-snapshot=ec2_utils.cli:volume_from_snapshot",
    "yaml-to-json=n_utils.cli:yaml_to_json",
    "yaml-to-yaml=n_utils.cli:yaml_to_yaml",
    "upsert-codebuild-projects=n_utils.cli:cli_upsert_codebuild_projects",
    "upsert-dns-record=n_utils.cli:upsert_dns_record",
    "update-sso-profile=n_utils.profile_util:cli_update_sso_profile",
    "deploy-connect=n_utils.cli:deploy_connect_contact_flows",
    "export-connect-contact-flow=n_utils.cli:export_connect_contact_flow",
    "list-connect-contact-flows=n_utils.cli:list_connect_contact_flows",
    "aws-config-to-json=n_utils.profile_util:cli_profiles_to_json",
]
NDT_ONLY_SCRIPT = [
    "bake-docker.sh",
    "bake-image.sh",
    "deploy-stack.sh",
    "undeploy-stack.sh",
    "deploy-serverless.sh",
    "undeploy-serverless.sh",
    "deploy-cdk.sh",
    "undeploy-cdk.sh",
    "deploy-terraform.sh",
    "undeploy-terraform.sh",
    "deploy-azure.sh",
    "undeploy-azure.sh",
    "print-create-instructions.sh",
    "terraform-pull-state.sh",
]
CONSOLE_ONLY = [
    "cf-update-stack=n_utils.cli:update_stack",
    "ndt=n_utils.ndt:ndt",
    "nameless-dt-register-complete=n_utils.project_util:ndt_register_complete",
    "nameless-dt-load-project-env=n_utils.project_util:load_project_env",
    "nameless-dt-enable-profile=n_utils.profile_util:cli_enable_profile",
]
CONSOLESCRIPTS = CONSOLE_ONLY + NDT_AND_CONSOLE
COMMAND_MAPPINGS = {}
for script in NDT_ONLY_SCRIPT:
    name = script
    value = "ndtscript"
    if name.endswith(".sh"):
        name = name[:-3]
        value = "ndtshell"
    COMMAND_MAPPINGS[name] = value
for script in NDT_AND_CONSOLE + NDT_ONLY:
    name, value = script.split("=")
    COMMAND_MAPPINGS[name] = value


def _to_str(data):
    ret = data
    decode_method = getattr(data, "decode", None)
    if callable(decode_method):
        try:
            ret = data.decode()
        except Exception:
            ret = _to_str(base64.b64encode(data))
    return str(ret)


def _to_bytes(data, encoding="utf-8"):
    ret = data
    encode_method = getattr(data, "encode", None)
    if callable(encode_method):
        ret = data.encode(encoding)
    return bytes(ret)



import json
from subprocess import PIPE, Popen

ARR_START = b"["
OBJ_START = b"{"


def az_login():
    proc = Popen(
        ["az", "login"],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    if proc.returncode != 0:
        return []
    return json.loads(output[output.find(ARR_START) :])


def az_logout():
    proc = Popen(
        ["az", "logout"],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    return proc.returncode == 0


def az_account_show():
    proc = Popen(
        ["az", "account", "show"],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, _err = proc.communicate()
    if proc.returncode != 0:
        return {}
    return json.loads(output[output.find(OBJ_START) :])


def az_account_list():
    proc = Popen(
        ["az", "account", "list"],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    if proc.returncode != 0:
        return []
    return json.loads(output[output.find(ARR_START) :])


def az_find_subscription_id(name, account_list):
    for subscription in account_list:
        if subscription["name"] == name or subscription["id"] == name:
            return subscription["id"]
    return None


def az_set_subscription(subscription_id):
    proc = Popen(
        ["az", "account", "set", "--subscription", subscription_id],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    return proc.returncode == 0


def az_select_subscription(name):
    curr_account = az_account_show()
    subscription_id = None
    if "name" in curr_account and "id" in curr_account:
        if curr_account["name"] == name or curr_account["id"] == name:
            return curr_account["id"]
        else:
            subscription_id = az_find_subscription_id(name, az_account_list())
    if not subscription_id:
        az_logout()
        subscription_id = az_find_subscription_id(name, az_login())

    if subscription_id:
        az_set_subscription(subscription_id)
    return subscription_id


def ensure_group(location, group_name):
    proc = Popen(
        ["az", "group", "list", "--query", "[?name=='" + group_name + "']"],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    groups = json.loads(output[output.find(ARR_START) :])
    if len(groups) == 0:
        proc = Popen(
            ["az", "group", "create", "--name", group_name, "--location", location],
            stdout=PIPE,
            stderr=PIPE,
        )
        output, err = proc.communicate()
        if proc.returncode == 0:
            groups = [json.loads(output[output.find(OBJ_START) :])]
        else:
            groups = [{}]
    return groups[0]


def ensure_management_group(group_name):
    proc = Popen(
        [
            "az",
            "account",
            "management-group",
            "list",
            "--query",
            "[?name=='" + group_name + "']",
        ],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    groups = json.loads(output[output.find(ARR_START) :])
    if len(groups) == 0:
        proc = Popen(
            ["az", "account", "management-group", "create", "--name", group_name],
            stdout=PIPE,
            stderr=PIPE,
        )
        output, err = proc.communicate()
        if proc.returncode == 0:
            groups = [json.loads(output[output.find(OBJ_START) :])]
        else:
            groups = [{}]


def delete_group(group_name):
    proc = Popen(
        ["az", "group", "delete", "-y", "--name", group_name],
        stdout=PIPE,
        stderr=PIPE,
    )
    output, err = proc.communicate()
    return proc.returncode == 0


def fetch_properties(parameters):
    cmd = get_az_command("show", parameters) + ["--query", "properties"]
    process = Popen(cmd, stdout=PIPE, stderr=PIPE)
    stdout, _ = process.communicate()
    if process.returncode != 0:
        return {}
    props = json.loads(stdout)
    ret = {}
    if "parameters" in props:
        for parameter, value in props["parameters"].items():
            if "value" in value:
                ret[parameter] = str(value["value"])
    if "outputs" in props:
        for output, value in props["outputs"].items():
            if "value" in value:
                ret[output] = str(value["value"])
    return ret


def get_az_command(command, parameters):
    scope = parameters.get("AZURE_SCOPE", "group")
    if "DEPLOYMENT_NAME" in parameters:
        name = parameters["DEPLOYMENT_NAME"]
    else:
        name = f"{parameters['BUILD_JOB_PREFIX']}_{parameters['COMPONENT']}_{parameters['ORIG_AZURE_NAME']}"
    ret = ["az", "deployment", command, "--name", name]
    if scope == "group":
        if "AZURE_GROUP" not in parameters:
            raise MissingAzureGroup
        ret.insert(2, "group")
        ret.append("--resource-group")
        ret.append(parameters["AZURE_GROUP"])
    elif scope == "management-group":
        if "AZURE_MANAGEMENT_GROUP" not in parameters:
            raise MissingAzureManagementGroup
        ret.insert(2, "mg")
        ret.append("--management-group-id")
        ret.append(parameters["AZURE_MANAGEMENT_GROUP"])
    elif scope == "subscription":
        ret.insert(2, "sub")
    elif scope == "tenant":
        ret.insert(2, "tenant")
    else:
        raise UnknownAzureGroupException
    return ret


class UnknownAzureGroupException(Exception):
    pass


class MissingAzureGroup(Exception):
    pass


class MissingAzureManagementGroup(Exception):
    pass



import os
from time import sleep, time

from threadlocal_aws import region, session
from threadlocal_aws.clients import cloudformation, organizations

from n_utils import cf_deploy, utils
from n_utils.ndt import find_include


def create_account(
    email,
    account_name,
    role_name="OrganizationAccountAccessRole",
    trust_role="TrustedAccountAccessRole",
    access_to_billing=True,
    trusted_accounts=None,
    mfa_token=None,
):
    if access_to_billing:
        access = "ALLOW"
    else:
        access = "DENY"
    timeout = 300

    if trusted_accounts:
        trusted_roles = {}
        for trusted_account in trusted_accounts:
            role_arn = find_role_arn(trusted_account)
            if not role_arn:
                raise Exception("Failed to resolve trusted account " + trusted_account)
            trusted_roles[trusted_account] = role_arn

    response = organizations().create_account(
        Email=email,
        AccountName=account_name,
        RoleName=role_name,
        IamUserAccessToBilling=access,
    )
    if "CreateAccountStatus" in response and "Id" in response["CreateAccountStatus"]:
        create_account_id = response["CreateAccountStatus"]["Id"]
        start_time = time()
        status = response["CreateAccountStatus"]["State"]
        while time() - start_time < timeout and not status == "SUCCEEDED":
            if response["CreateAccountStatus"]["State"] == "FAILED":
                raise Exception("Account creation failed: " + response["CreateAccountStatus"]["FailureReason"])
            print("Waiting for account creation to finish")
            sleep(2)
            response = organizations().describe_create_account_status(CreateAccountRequestId=create_account_id)
            status = response["CreateAccountStatus"]["State"]
        if time() - start_time > timeout and not status == "SUCCEEDED":
            raise Exception("Timed out waiting to create account " + response["CreateAccountStatus"]["State"])
        account_id = response["CreateAccountStatus"]["AccountId"]

    os.environ["paramManagedAccount"] = account_id
    os.environ["paramManageRole"] = role_name
    template = find_include("manage-account.yaml")
    cf_deploy.deploy("managed-account-" + account_name, template, region())

    if trusted_accounts:
        role_arn = "arn:aws:iam::" + account_id + ":role/" + role_name
        assumed_creds = utils.assume_role(role_arn, mfa_token_name=mfa_token)
        sess = session(
            aws_access_key_id=assumed_creds["AccessKeyId"],
            aws_secret_access_key=assumed_creds["SecretAccessKey"],
            aws_session_token=assumed_creds["SessionToken"],
        )
        for trusted_account in trusted_accounts:
            os.environ["paramTrustedAccount"] = trusted_roles[trusted_account].split(":")[4]
            os.environ["paramRoleName"] = trust_role
            template = find_include("trust-account-role.yaml")
            cf_deploy.deploy("trust-" + trusted_account, template, utils.region(), session=sess)
        template = find_include("manage-account.yaml")
        for trusted_account in trusted_accounts:
            role_arn = trusted_roles[trusted_account]
            print("Assuming role " + role_arn)
            assumed_creds = utils.assume_role(role_arn, mfa_token_name=mfa_token)
            sess = session(
                aws_access_key_id=assumed_creds["AccessKeyId"],
                aws_secret_access_key=assumed_creds["SecretAccessKey"],
                aws_session_token=assumed_creds["SessionToken"],
            )
            os.environ["paramManagedAccount"] = account_id
            os.environ["paramRoleName"] = trust_role
            cf_deploy.deploy(
                "managed-account-" + account_name + "-" + account_id,
                template,
                region(),
                session=sess,
            )


def find_role_arn(trusted_account):
    cf_stacks = cloudformation().get_paginator("describe_stacks")
    for page in cf_stacks.paginate():
        for stack in page["Stacks"]:
            if stack["StackName"].endswith(trusted_account) or "-" + trusted_account + "-" in stack["StackName"]:
                for output in stack["Outputs"]:
                    if output["OutputKey"] == "ManageRole":
                        return output["OutputValue"]
    return None


def list_created_accounts():
    cf_stacks = cloudformation().get_paginator("describe_stacks")
    for page in cf_stacks.paginate():
        for stack in page["Stacks"]:
            if stack["StackName"].startswith("managed-account-"):
                yield "-".join(stack["StackName"].split("-")[2:-1])



#!/usr/bin/env python

import base64
import json
import os

# Copyright 2017 Nitor Creations Oy
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
from os import walk

import pyotp
import pyqrcode
import yaml
from Cryptodome.Cipher import AES
from Cryptodome.Hash import SHA256
from Cryptodome.Util import Counter

from n_utils import _to_bytes, _to_str
from n_utils.bw_util import get_bwentry
from n_utils.yuuuu3332111i1l1i import I11iIi1I, IiII1IiiIiI1


def dthandler(obj):
    return obj.isoformat() if hasattr(obj, "isoformat") else json.JSONEncoder().default(obj)


def mfa_add_token(args):
    """Adds or overwrites an MFA token to be used with role assumption.
    Tokens will be saved in a .ndt subdirectory in the user's home directory."""
    ndt_dir = get_ndt_dir()
    if not os.path.exists(ndt_dir):
        os.makedirs(ndt_dir)
    data = {
        "token_name": args.token_name,
    }
    if args.token_arn:
        data["token_arn"] = args.token_arn
    if args.bitwarden_entry:
        data["bitwarden_entry"] = args.bitwarden_entry
    else:
        data["token_secret"] = "enc--" + _to_str(IiII1IiiIiI1(_to_bytes(args.token_secret)))
    token_file = ndt_dir + "/mfa_" + args.token_name
    if os.path.isfile(token_file) and not args.force:
        raise ValueError("A token with the name " + args.token_name + " already exists!")
    with open(token_file, "w") as outfile:
        os.chmod(token_file, 0o600)
        yaml.safe_dump(data, outfile, allow_unicode=True, default_flow_style=False)


def mfa_read_token(token_name):
    """Reads a previously added MFA token file and returns its data."""
    data = None
    with open(get_ndt_dir() + "/mfa_" + token_name) as infile:
        try:
            data = yaml.load(infile, yaml.SafeLoader)
        except yaml.YAMLError as exc:
            print(exc)
    if data:
        if "bitwarden_entry" in data and data["bitwarden_entry"]:
            bw_entry = get_bwentry(data["bitwarden_entry"])
            if "token_arn" in bw_entry.fields:
                data["token_arn"] = bw_entry.fields["token_arn"]
            if bw_entry.totp:
                data["token_secret"] = "enc--" + _to_str(IiII1IiiIiI1(_to_bytes(bw_entry.totp)))
            else:
                raise ValueError("No totp secret found for bitwarden entry '" + data["bitwarden_entry"] + "'")
        if "token_secret" in data and not data["token_secret"].startswith("enc--"):
            data["force"] = True
            mfa_add_token(Struct(**data))
            return mfa_read_token(token_name)
    return data


def get_ndt_dir():
    """Gets cross platform ndt directory path. Makes sure it exists."""
    ndt_dir = os.path.expanduser("~/.ndt")
    if not os.path.exists(ndt_dir):
        os.makedirs(ndt_dir)
    return ndt_dir


def mfa_generate_code(token_name):
    """Generates an MFA code with the specified token."""
    token = mfa_read_token(token_name)
    if token["token_secret"].startswith("enc--"):
        secret = I11iIi1I(token["token_secret"][5:])
    else:
        secret = token["token_secret"]
    totp = pyotp.TOTP(secret)
    return totp.now()


def mfa_to_qrcode(token_name):
    """Generates a qr code of the token for importing into other devices"""
    token = mfa_read_token(token_name)
    if token["token_secret"].startswith("enc--"):
        secret = I11iIi1I(token["token_secret"][5:])
    else:
        secret = token["token_secret"]
    url = "otpauth://totp/" + token_name + "?secret=" + secret
    qr = pyqrcode.create(url)
    print(qr.terminal())


def mfa_generate_code_with_secret(secret):
    """Generates an MFA code using the secret passed in."""
    if _to_str(secret).startswith("enc--"):
        secret = I11iIi1I(secret[5:])
    totp = pyotp.TOTP(secret)
    return totp.now()


def mfa_delete_token(token_name):
    """Deletes an MFA token file from the .ndt subdirectory in the user's
    home directory"""
    os.remove(get_ndt_dir() + "/mfa_" + token_name)


def mfa_backup_tokens(backup_secret):
    """Writes MFA secrets encrypted with backup_secret and base64 encoded to stdout."""
    tokens = []
    for token in list_mfa_tokens():
        token_data = mfa_read_token(token)
        if token_data["token_secret"].startswith("enc--"):
            token_data["token_secret"] = _to_str(I11iIi1I(token_data["token_secret"][5:]))
        tokens.append(token_data)
    counter = Counter.new(128, initial_value=1337)
    cipher = AES.new(get_backup_key_digest(backup_secret), AES.MODE_CTR, counter=counter)
    return _to_str(base64.b64encode(cipher.encrypt(_to_bytes(json.dumps(tokens, default=dthandler)))))


def mfa_decrypt_backup_tokens(backup_secret, file):
    """Decrypts backed up MFA secrets from file, prints to stdout."""
    with open(os.path.expanduser(file)) as infile:
        data = infile.read()
    counter = Counter.new(128, initial_value=1337)
    cipher = AES.new(get_backup_key_digest(backup_secret), AES.MODE_CTR, counter=counter)
    return cipher.decrypt(base64.b64decode(data)).decode()


class Struct:
    def __init__(self, **entries):
        self.__dict__.update(entries)


def list_mfa_tokens():
    tokens = []
    for dirpath, dirnames, filenames in walk(get_ndt_dir()):
        tokens.extend([fn[4:] for fn in filenames if fn.startswith("mfa_")])
        break
    return tokens


def get_backup_key_digest(backup_secret):
    key = SHA256.new()
    key.update(backup_secret.encode("utf-8"))
    return key.digest()



import time

from threadlocal_aws.clients import route53


def upsert_record(dns_name, record_type, value, ttl=300, wait=True):
    zone_id, change = get_record_change(dns_name, record_type, value, hosted_zones(), ttl=ttl)
    request = route53().change_resource_record_sets(HostedZoneId=zone_id, ChangeBatch={"Changes": [change]})[
        "ChangeInfo"
    ]
    if wait:
        status = route53().get_change(Id=request["Id"])["ChangeInfo"]["Status"]
        while not status == "INSYNC":
            print("Waiting for request to sync - " + str(request["Id"]) + "(" + dns_name + ") status: " + status)
            time.sleep(2)
            status = route53().get_change(Id=request["Id"])["ChangeInfo"]["Status"]


def hosted_zones():
    pages = route53().get_paginator("list_hosted_zones")
    for page in pages.paginate():
        yield from page.get("HostedZones", [])


def longest_matching_zone(alias, hosted_zones):
    ret = {"Name": ""}
    for zone in hosted_zones:
        if (alias + ".").endswith(zone["Name"]) and len(zone["Name"]) > len(ret["Name"]):
            ret = zone
    return ret


def get_record_change(dns_name, record_type, value, hosted_zones, ttl=300):
    zone = longest_matching_zone(dns_name, hosted_zones)
    if zone:
        print(dns_name + " => " + "(" + record_type + ") '" + value + "' in " + zone["Name"])
        change = {
            "Action": "UPSERT",
            "ResourceRecordSet": {
                "Name": dns_name,
                "Type": record_type,
                "TTL": ttl,
                "ResourceRecords": [{"Value": value}],
            },
        }
        return zone["Id"], change



#!/usr/bin/env python

# Copyright 2016-2017 Nitor Creations Oy
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Utilities to bootsrap AWS accounts into use with nameless-deploy-tools
"""

import argparse
import os
import random
import re
import shutil
import stat
import subprocess
import sys
from ast import literal_eval
from collections import OrderedDict
from copy import deepcopy
from subprocess import PIPE, Popen

import argcomplete
import ipaddr
from argcomplete.completers import ChoicesCompleter
from threadlocal_aws.clients import ec2, route53

from n_utils import _to_str
from n_utils.aws_infra_util import yaml_load, yaml_save
from n_utils.ndt import find_all_includes, find_include
from n_utils.utils import has_output_selector, select_stacks
from n_utils.writer import ConfigFileWriter


def enum(**enums):
    return type("Enum", (), enums)


def type_guess(val):
    if not val:
        return None
    try:
        return literal_eval(val)
    except Exception:
        return _to_str(val)


BRANCH_MODES = enum(SINGLE_STACK="single", MULTI_STACK="multi")


def _add_default_params(parser):
    parser.add_argument("template", nargs="?").completer = ChoicesCompleter(list_templates())
    parser.add_argument(
        "-y",
        "--yes",
        action="store_true",
        help="Answer yes or use default to all questions",
    )


def create_stack():
    """Create a stack from a template"""
    parser = argparse.ArgumentParser(description=create_stack.__doc__, add_help=False)
    _add_default_params(parser)
    parser.add_argument("-h", "--help", action="store_true")
    if "_ARGCOMPLETE" in os.environ:
        ac_args = argcomplete.split_line(os.environ["COMP_LINE"], os.environ["COMP_POINT"])[3]
        if len(ac_args) >= 2:
            template = ac_args[1]
            template_yaml = load_template(template)
            if template_yaml and "ContextClass" in template_yaml:
                context = load_class(template_yaml["ContextClass"])()
                context.add_context_arguments(parser)
    argcomplete.autocomplete(parser)
    args, _ = parser.parse_known_args()
    if args.template:
        template_yaml = load_template(args.template)
        if "ContextClass" in template_yaml:
            context = load_class(template_yaml["ContextClass"])()
            template_yaml.pop("ContextClass", None)
            parser = argparse.ArgumentParser(description=context.__doc__)
            _add_default_params(parser)
            context.add_context_arguments(parser)
        else:
            parser = argparse.ArgumentParser(description=create_stack.__doc__)
            _add_default_params(parser)
    else:
        parser = argparse.ArgumentParser(description=create_stack.__doc__)
        _add_default_params(parser)
    args = parser.parse_args()
    context.resolve_parameters(args)
    context.set_template(template_yaml)
    if context.write(yes=args.yes):
        subprocess.check_call(
            [
                "ndt",
                "print-create-instructions",
                context.component_name,
                context.stack_name,
            ]
        )
    return


def load_template(template):
    file_name = find_include("creatable-templates/" + template + ".yaml")
    if file_name:
        return yaml_load(open(file_name))
    else:
        return None


def load_class(name):
    components = name.split(".")
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def list_templates():
    ret = []
    files = find_all_includes("creatable-templates/*.yaml")
    for next_file in files:
        next_name = os.path.basename(next_file)[:-5]
        if next_name not in ret:
            ret.append(next_name)
    return ret


def has_entry(prefix, name, file_name):
    if not os.path.isfile(file_name):
        return False
    with open(file_name) as config:
        for line in config.readlines():
            if "[" + prefix + name + "]" == line.strip():
                return True
    return False


def setup_cli(name=None, key_id=None, secret=None, region=None):
    if name is None:
        name = type_guess(input("Profile name: "))
    home_dir = os.path.expanduser("~")
    config_file = os.path.join(home_dir, ".aws", "config")
    credentials_file = os.path.join(home_dir, ".aws", "credentials")
    if has_entry("profile ", name, config_file) or has_entry("", name, credentials_file):
        print("Profile " + name + " already exists. Not overwriting.")
        return
    if key_id is None:
        key_id = type_guess(input("Key ID: "))
    if secret is None:
        secret = type_guess(input("Key secret: "))
    if region is None:
        region = type_guess(input("Default region: "))
    writer = ConfigFileWriter()
    config_values = {
        "__section__": "profile " + name,
        "output": "json",
        "region": region,
    }
    credentials_values = {
        "__section__": name,
        "aws_access_key_id": key_id,
        "aws_secret_access_key": secret,
    }
    writer.update_config(config_values, config_file)
    writer.update_config(credentials_values, credentials_file)
    home_bin = credentials_file = os.path.join(home_dir, "bin")
    if not os.path.isdir(home_bin):
        os.makedirs(home_bin)
    source_file = os.path.join(home_bin, name)
    with open(source_file, "w") as source_script:
        source_script.write("#!/bin/bash\n\n")
        source_script.write("export AWS_DEFAULT_REGION=")
        source_script.write(region + " AWS_PROFILE=" + name)
        source_script.write(" AWS_DEFAULT_PROFILE=" + name + "\n")
    os.chmod(source_file, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
    os.environ["AWS_PROFILE"] = name
    os.environ["AWS_DEFAULT_PROFILE"] = name
    os.environ["AWS_DEFAULT_REGION"] = region


def _append_cidr_param(public, letter, cidr, params):
    if public:
        cidr_param = "paramPub" + letter + "Cidr"
        description = "Public Subnet " + letter + " CIDR block"
    else:
        cidr_param = "paramPriv" + letter + "Cidr"
        description = "Private Subnet " + letter + " CIDR block"
    params.update(
        OrderedDict(
            [
                (
                    cidr_param,
                    OrderedDict(
                        [
                            ("Description", description),
                            ("Type", "String"),
                            ("Default", _nts(cidr)),
                        ]
                    ),
                )
            ]
        )
    )


def _append_network_resources(public, letter, resources, availability_zone):
    if public:
        subnet_resource = "resourcePubSubnetA"
        route_table_assoc = "resourcePubSubnetRouteTableAssocA"
        cidr_param = "paramPub" + letter + "Cidr"
        db_subnet = "resourcePubSubnetGroup"
    else:
        subnet_resource = "resourcePrivSubnetA"
        route_table_assoc = "resourcePrivSubnetRouteTableAssocA"
        cidr_param = "paramPriv" + letter + "Cidr"
        route_table = "resourcePrivRouteTableA"
        db_subnet = "resourcePrivSubnetGroup"

    subnet_to_create = subnet_resource[:-1] + letter

    subnet_res_dict = deepcopy(resources[subnet_resource])
    subnet_res_dict["Properties"]["CidrBlock"]["Ref"] = cidr_param
    subnet_res_dict["Properties"]["AvailabilityZone"] = availability_zone
    subnet_res_dict["Properties"]["Tags"][0]["Value"]["Fn::Join"][1][1] += availability_zone
    resources.update({subnet_to_create: subnet_res_dict})
    if not public:
        route_table_dict = deepcopy(resources[route_table])
        route_table_dict["Properties"]["Tags"][0]["Value"]["Fn::Join"][1][1] = " private route table " + letter
        resources.update(OrderedDict([(route_table[:-1] + letter, route_table_dict)]))
    route_table_assoc_dict = deepcopy(resources[route_table_assoc])
    if not public:
        route_table_assoc_dict["Properties"]["RouteTableId"]["Ref"] = route_table[:-1] + letter
    route_table_assoc_dict["Properties"]["SubnetId"]["Ref"] = subnet_to_create
    resources.update(OrderedDict([(route_table_assoc[:-1] + letter, route_table_assoc_dict)]))

    resources[db_subnet]["Properties"]["SubnetIds"].append(OrderedDict([("Ref", subnet_to_create)]))


def _nts(net):
    return str(net).lower()


def _get_network_yaml(network, vpc_cidr, subnet_prefixlen, subnet_base, network_yaml, common_yaml):
    subnet_bits = 32 - int(subnet_prefixlen)
    subnet_size = 2**subnet_bits
    last_subnet = subnet_base - subnet_size
    az_response = ec2().describe_availability_zones()
    az_names = sorted([az_data["ZoneName"] for az_data in az_response["AvailabilityZones"]])
    network_yaml["Parameters"]["paramVPCCidr"]["Default"] = _nts(vpc_cidr)
    for az_name in az_names:
        zone_letter = az_name[-1:]
        zone_upper_letter = zone_letter.upper()
        last_subnet += subnet_size
        private_subnet_addr = last_subnet + (len(az_names) * subnet_size)
        subnet = ipaddr.IPv4Network(str(last_subnet) + "/" + str(subnet_prefixlen))
        private_subnet = ipaddr.IPv4Network(str(private_subnet_addr) + "/" + str(subnet_prefixlen))
        if zone_letter == "a":
            network_yaml["Parameters"]["paramPubACidr"]["Default"] = _nts(subnet)
            network_yaml["Parameters"]["paramPrivACidr"]["Default"] = _nts(private_subnet)
            network_yaml["Resources"]["resourcePubSubnetA"]["Properties"]["AvailabilityZone"] = az_name
            network_yaml["Resources"]["resourcePrivSubnetA"]["Properties"]["AvailabilityZone"] = az_name
            common_yaml["paramNetwork"]["Default"] = _nts(network)
        else:
            _append_cidr_param(True, zone_upper_letter, _nts(subnet), network_yaml["Parameters"])
            _append_cidr_param(
                False,
                zone_upper_letter,
                _nts(private_subnet),
                network_yaml["Parameters"],
            )
            _append_network_resources(True, zone_upper_letter, network_yaml["Resources"], az_name)
            _append_network_resources(False, zone_upper_letter, network_yaml["Resources"], az_name)
            network_yaml["Outputs"]["subnet" + zone_upper_letter] = {
                "Description": "Public Subnet " + zone_upper_letter,
                "Value": {"Ref": "resourcePubSubnet" + zone_upper_letter},
                "Export": {
                    "Name": {
                        "Fn::Join": [
                            ":",
                            [
                                {"Ref": "AWS::StackName"},
                                "publicSubnet" + zone_upper_letter,
                            ],
                        ]
                    }
                },
            }
            network_yaml["Outputs"]["subnetPriv" + zone_upper_letter] = {
                "Description": "Private Subnet " + zone_upper_letter,
                "Value": {"Ref": "resourcePrivSubnet" + zone_upper_letter},
                "Export": {
                    "Name": {
                        "Fn::Join": [
                            ":",
                            [
                                {"Ref": "AWS::StackName"},
                                "privateSubnet" + zone_upper_letter,
                            ],
                        ]
                    }
                },
            }
            pub_net = deepcopy(common_yaml["paramSubnetA"])
            pub_net["Description"] = pub_net["Description"][:-1] + zone_upper_letter
            pub_net["Default"]["StackRef"]["paramName"] = (
                pub_net["Default"]["StackRef"]["paramName"][:-1] + zone_upper_letter
            )
            common_yaml["paramSubnet" + zone_upper_letter] = pub_net
            priv_net = deepcopy(common_yaml["paramSubnetPrivA"])
            priv_net["Description"] = pub_net["Description"][:-1] + zone_upper_letter
            priv_net["Default"]["StackRef"]["paramName"] = (
                priv_net["Default"]["StackRef"]["paramName"][:-1] + zone_upper_letter
            )
            common_yaml["paramSubnetPriv" + zone_upper_letter] = priv_net
    return network_yaml, common_yaml


def _set_first_parameter(root, name, value):
    if name in root:
        root[name]["Default"] = value
        return True
    for k, v in list(root.items()):
        if isinstance(v, dict) or isinstance(v, OrderedDict):
            return _set_first_parameter(v, name, value)
        elif isinstance(v, list):
            for next in v:
                if _set_first_parameter(next, name, value):
                    return True


def _map_ssh_key(context, param, value):
    key_name = None
    default_name = "ndt-" + context.__class__.__name__.lower() + "-instance"
    if value == "1":
        key_name = type_guess(input("Name for new key pair (" + default_name + "): "))
        if not key_name:
            key_name = default_name
        key_res = ec2().create_key_pair(KeyName=key_name)
        if key_res and "KeyMaterial" in key_res:
            proc = Popen(
                ["store-secret.sh", key_name + ".pem"],
                stdout=PIPE,
                stdin=PIPE,
                stderr=PIPE,
            )
            proc.communicate(input=key_res["KeyMaterial"])
    else:
        try:
            index = int(value) - 2
            key_name = context.ssh_keys[index]
        except (ValueError, IndexError):
            print("Invalid ssh key selection " + value)
            sys.exit(1)
    return key_name


def _map_elastic_ip(context, param, value):
    eip = None
    if value == "1":
        eip_res = ec2().allocate_address(Domain="vpc")
        if eip_res and "PublicIp" in eip_res:
            eip = eip_res["PublicIp"]
        ec2().create_tags(
            Resources=[eip_res["AllocationId"]],
            Tags=[{"Key": "Name", "Value": context.__class__.__name__.lower()}],
        )
    else:
        try:
            index = int(value) - 2
            eip = context.elastic_ips[index]
        except (ValueError, IndexError):
            print("Invalid elastic ip selection " + value)
            sys.exit(1)
    return eip


def _map_list(context, param, value):
    select_list = getattr(context, param + "s")
    try:
        index = int(value) - 1
        return select_list[index]
    except (ValueError, IndexError):
        print("Invalid " + param + " selection " + value)
        sys.exit(1)


class ContextClassBase:
    """Base class for template contexts. Can be directly used for
    bootstrap stacks with no parameters to ask can use it directly
    """

    ask_fields = []
    component_name = "bootstrap"
    stack_name = "Stack name ({0}): "
    branch_mode = BRANCH_MODES.SINGLE_STACK
    template = None
    value_mappers = {}
    template_transformers = []

    def __init__(self, ask_fields):
        if not ask_fields:
            ask_fields = ["stack_name"]
        elif "stack_name" not in ask_fields:
            if ask_fields[0] == "component_name":
                ask_fields.insert(1, "stack_name")
            else:
                ask_fields.insert(0, "stack_name")
        self.ask_fields = ask_fields

    def stack_name_default(self):
        return "default"

    def set_template(self, template):
        self.template = template
        for transformer in self.template_transformers:
            transformer(self)

    def format_prompt(self, parameter, default=None):
        prompt = getattr(self, parameter)
        if not default:
            default = getattr(self, parameter + "_default")()
        return prompt.format(default)

    def getattr(self, parameter):
        return getattr(self, parameter)

    def setattr(self, parameter, value):
        return setattr(self, parameter, value)

    def add_context_arguments(self, parser):
        parameters = self.getattr("ask_fields")
        shorts = ["h"]
        for param in parameters:
            if param[0] not in shorts:
                shorts.append(param[0])
            else:
                attempt = [x[0] for x in re.split("[\b_]", param)]
                if len(attempt) > 1 and attempt[1] not in shorts:
                    shorts.append(attempt[1])
                else:
                    attempt = param[1]
                    if attempt not in shorts:
                        shorts.append(attempt)
                    else:
                        shorts.append("")
        shorts.pop(0)
        for short_arg, long_arg in zip(shorts, parameters):
            if short_arg:
                parser.add_argument("-" + short_arg, "--" + long_arg, help=self.format_prompt(long_arg))
            else:
                parser.add_argument("--" + long_arg, help=self.format_prompt(long_arg))

    def resolve_parameters(self, args):
        for param in self.ask_fields:
            if getattr(args, param):
                setattr(self, param, getattr(args, param))
            elif args.yes:
                setattr(self, param, getattr(args, param + "_default")())
            else:
                default = self.getattr(param + "_default")()
                prompt = self.format_prompt(param, default=default)
                setval = type_guess(input(prompt))
                if not setval:
                    setval = default
                if param in self.value_mappers:
                    setval = self.value_mappers[param](self, param, setval)
                self.setattr(param, setval)

    def _ask_ssh_key(self):
        self.ssh_key = "Ssh key ({0}):\n1: create new\n"
        self.ssh_keys = []
        self.ssh_key_default = lambda: "1"
        if "ssh_key" not in self.ask_fields:
            self.ask_fields.append("ssh_key")
        index = 2
        keys = ec2().describe_key_pairs()
        if keys and "KeyPairs" in keys:
            for key in keys["KeyPairs"]:
                self.ssh_keys.append(key["KeyName"])
                self.ssh_key = self.ssh_key + str(index) + ": " + key["KeyName"] + "\n"
                index = index + 1
        self.value_mappers["ssh_key"] = _map_ssh_key
        self.template_transformers.append(
            lambda myself: _set_first_parameter(myself.template, "paramSshKeyName", myself.ssh_key)
        )

    def _ask_elastic_ip(self):
        self.elastic_ip = "Elastic ip ({0}):\n1: allocate new\n"
        self.elastic_ips = []
        self.elastic_ip_default = lambda: "1"
        if "elastic_ip" not in self.ask_fields:
            self.ask_fields.append("elastic_ip")
        index = 2
        eips = ec2().describe_addresses()
        if eips and "Addresses" in eips:
            for address in eips["Addresses"]:
                if "InstanceId" not in address:
                    self.elastic_ips.append(address["PublicIp"])
                    self.elastic_ip = self.elastic_ip + str(index) + ": " + address["PublicIp"] + "\n"
                    index = index + 1
        self.value_mappers["elastic_ip"] = _map_elastic_ip
        self.template_transformers.append(
            lambda myself: _set_first_parameter(myself.template, "paramEip", myself.elastic_ip)
        )

    def write(self, yes=False):
        if "Files" in self.template:
            for entry in self.template["Files"]:
                for source, dest in list(entry.items()):
                    dest = self.component_name + os.sep + dest % self.__dict__
                    dest = os.path.normpath(dest)
                    dest_dir = os.path.normpath(os.path.dirname(dest))
                    if not os.path.exists(dest_dir):
                        os.makedirs(dest_dir)
                    source_file = find_include(source)
                    shutil.copy2(source_file, dest)
            self.template.pop("Files", None)
        stack_dir = os.path.join(".", self.component_name, "stack-" + self.stack_name)
        if not os.path.exists(stack_dir):
            os.makedirs(stack_dir)
        if self.branch_mode == BRANCH_MODES.SINGLE_STACK:
            file_name = "infra.properties"
            stack_props = os.path.join(stack_dir, file_name)
            if not os.path.exists(stack_props):
                with open(stack_props, "w") as stack_props_file:
                    stack_props_file.write("STACK_NAME=$ORIG_STACK_NAME\n")
        stack_template = os.path.join(stack_dir, "template.yaml")
        if os.path.exists(stack_template) and not yes:
            answer = type_guess(input("Overwrite " + self.stack_name + " stack? (n): "))
            if not answer or not answer.lower() == "y":
                return False
        with open(stack_template, "w") as stack_file:
            stack_file.write(yaml_save(self.template))
        return True


class Network(ContextClassBase):
    """Creates a public and private subnet for every availability zone
    in the selected region and a shared common yaml (common/network.yaml)
    for easy access to parameters.
    """

    stack_name = "Network stack name ({0}): "
    vpc_cidr = "VPC CIDR ({0}) "
    subnet_prefixlen = "Subnet prefix length ({0}) "
    subnet_base = "Subnet base ({0}) "
    last_vpc_cidr = None
    common_yaml = None

    def __init__(self):
        ContextClassBase.__init__(self, ["vpc_cidr", "subnet_prefixlen", "subnet_base"])

    def stack_name_default(self):
        return "network"

    def vpc_cidr_default(self):
        self.last_vpc_cidr = "10." + str(random.randint(0, 255)) + ".0.0/16"
        return self.last_vpc_cidr

    def subnet_prefixlen_default(self):
        try:
            nw = ipaddr.IPv4Network(self.vpc_cidr)
            return nw.prefixlen + 4
        except BaseException:
            try:
                nw = ipaddr.IPv4Network(self.last_vpc_cidr)
                return nw.prefixlen + 4
            except BaseException:
                return 20

    def subnet_base_default(self):
        try:
            nw = ipaddr.IPv4Network(self.vpc_cidr)
            return nw.network
        except BaseException:
            try:
                nw = ipaddr.IPv4Network(self.last_vpc_cidr)
                return nw.network
            except BaseException:
                return "10." + str(random.randint(0, 255)) + ".0.0"

    def set_template(self, template):
        common_yaml = yaml_load(open(find_include("creatable-templates/network/common.yaml")))
        self.template, self.common_yaml = _get_network_yaml(
            self.stack_name,
            self.vpc_cidr,
            self.subnet_prefixlen,
            self.subnet_base,
            template,
            common_yaml,
        )

    def write(self, yes=False):
        ContextClassBase.write(self, yes=yes)
        if not os.path.exists("common"):
            os.makedirs("common")
        common_out = os.path.join("common", "network.yaml")
        with open(common_out, "w") as c_out:
            c_out.write(yaml_save(self.common_yaml))
        return True


class BakeryRoles(ContextClassBase):
    """Creates roles necessary for baking images and deploying stacks"""

    def __init__(self):
        ContextClassBase.__init__(self, ["network_stack", "vault_stack"])
        self._ask_network_stack()
        self._ask_vault_stack()

    def _ask_network_stack(self):
        self.network_stack = "Network stack ({0}):\n"

        def network_sel(stack):
            return has_output_selector(stack, "VPC", lambda stack: stack["StackName"])

        self.network_stacks = select_stacks(network_sel)
        if self.network_stacks:
            index = 1
            for stack_name in self.network_stacks:
                self.network_stack = self.network_stack + str(index) + ": " + stack_name + "\n"
                index = index + 1
        self.value_mappers["network_stack"] = _map_list
        self.template_transformers.append(
            lambda myself: _set_first_parameter(myself.template, "paramNetworkStack", myself.network_stack)
        )

    def _ask_vault_stack(self):
        self.vault_stack = "Vault stack ({0}):\n"

        def vault_sel(stack):
            return has_output_selector(stack, "decryptPolicy", lambda stack: stack["StackName"])

        self.vault_stacks = select_stacks(vault_sel)
        if self.vault_stacks:
            index = 1
            for stack_name in self.vault_stacks:
                self.vault_stack = self.vault_stack + str(index) + ": " + stack_name + "\n"
                index = index + 1
        self.value_mappers["vault_stack"] = _map_list
        self.template_transformers.append(
            lambda myself: _set_first_parameter(myself.template, "paramVaultStack", myself.vault_stack)
        )

    def stack_name_default(self):
        return "bakery-roles"

    def network_stack_default(self):
        return "1"

    def vault_stack_default(self):
        return "1"


class Route53(ContextClassBase):
    """Creates a common shared yaml (common/route53.yaml) for
    easy access to route53 parameters
    """

    def __init__(self):
        ContextClassBase.__init__(self, ["hosted_zone"])
        self.ask_fields.pop(0)
        self._ask_hosted_zone()

    def _ask_hosted_zone(self):
        self.hosted_zone = "Hosted zone ({0}):\n"
        self.hosted_zones = route53().list_hosted_zones()["HostedZones"]
        self.hosted_zone_default = lambda: "1"
        if self.hosted_zones:
            index = 1
            for zone in self.hosted_zones:
                priv = "Private"
                if zone["Config"]["PrivateZone"]:
                    priv = "Public"
                self.hosted_zone = (
                    self.hosted_zone
                    + str(index)
                    + ": "
                    + zone["Name"]
                    + " ("
                    + zone["Id"].split("/")[-1:][0]
                    + ") - "
                    + priv
                    + "\n"
                )
                index = index + 1
        self.value_mappers["hosted_zone"] = _map_list

    def write(self, yes=False):
        if not os.path.exists("common"):
            os.makedirs("common")
        common_out = os.path.join("common", "route53.yaml")
        common_yaml = {
            "paramHostedZoneName": {
                "Type": "String",
                "Description": "Name of hosted zone to use",
                "Default": self.hosted_zone["Name"],
            },
            "paramHostedZoneDomain": {
                "Type": "String",
                "Description": "Name of hosted zone to use",
                "Default": self.hosted_zone["Name"][:-1],
            },
            "paramHostedZoneId": {
                "Type": "String",
                "Description": "Id of hosted zone to use",
                "Default": self.hosted_zone["Id"].split("/")[-1:][0],
            },
        }
        with open(common_out, "w") as c_out:
            c_out.write(yaml_save(common_yaml))
        return False


class Jenkins(ContextClassBase):
    """Creates a jenkins stack with the default domain name jenkins.${paramHostedZoneDomain}.
    Also a Jenkins Job DSL script is created that can be run in Jenkins to create jobs
    to bake and deploy ndt infrastructure defined in this repository.
    """

    elastic_ip = "Elastic IP ({0}):\n1: allocate new\n"
    elastic_ips = []

    def __init__(self):
        ContextClassBase.__init__(self, ["component_name", "elastic_ip", "ssh_key"])
        self.component_name = "Component name ({0}): "
        self.branch_mode = BRANCH_MODES.MULTI_STACK
        self._ask_elastic_ip()
        self._ask_ssh_key()

    def component_name_default(self):
        return "jenkins"

    def stack_name_default(self):
        return "jenkins"

    def set_template(self, template):
        ContextClassBase.set_template(self, template)
        az_response = ec2().describe_availability_zones()
        az_names = sorted([az_data["ZoneName"] for az_data in az_response["AvailabilityZones"]])
        for az_name in az_names:
            self.template["Resources"]["Fn::Merge"][0]["resourceAsg"]["Properties"]["AvailabilityZones"].append(az_name)
            self.template["Resources"]["Fn::Merge"][0]["resourceAsg"]["Properties"]["VPCZoneIdentifier"].append(
                "paramSubnet" + az_name[-1:].upper()
            )



#!/usr/bin/env python

# Copyright 2016-2017 Nitor Creations Oy
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import hashlib
import locale
import os
import re
import sys
import time
from datetime import datetime

from botocore.exceptions import ClientError
from ec2_utils.logs import CloudWatchLogsThread, fmttime
from pygments import formatters, highlight, lexers
from pygments.styles import get_style_by_name
from termcolor import colored
from threadlocal_aws.clients import cloudformation, ec2, s3

from n_utils import aws_infra_util
from n_utils.log_events import CloudFormationEvents
from n_utils.utils import get_images

REDIRECTED = False


def log_data(data, output_format="yaml"):
    if output_format == "yaml":
        formatted = aws_infra_util.yaml_save(data)
    else:
        formatted = aws_infra_util.json_save(data)
    lexer = lexers.get_lexer_by_name(output_format)
    formatter = formatters.get_formatter_by_name("256")
    formatter.__init__(style=get_style_by_name("emacs"))
    if not isinstance(formatted, str):
        formatted = str(formatted, "UTF-8")
    colored_yaml = os.linesep + highlight(formatted, lexer, formatter)
    log(colored_yaml)


def log(message):
    os.write(
        1,
        (colored(fmttime(datetime.now()), "yellow") + " " + message + os.linesep).encode(locale.getpreferredencoding()),
    )


def update_stack(stack_name, template, params, dry_run=False, session=None, tags=None, disable_rollback=False):
    clf = cloudformation(session=session)
    chset_name = stack_name + "-" + time.strftime("%Y%m%d%H%M%S", time.gmtime())
    params = get_template_arguments(stack_name, template, params)
    params["ChangeSetName"] = chset_name
    if tags:
        params["Tags"] = tags
    chset_id = clf.create_change_set(**params)["Id"]
    chset_data = clf.describe_change_set(ChangeSetName=chset_id)
    status = chset_data["Status"]
    while "_COMPLETE" not in status and status != "FAILED":
        time.sleep(5)
        chset_data = clf.describe_change_set(ChangeSetName=chset_id)
        status = chset_data["Status"]
    if status == "FAILED":
        clf.delete_change_set(ChangeSetName=chset_id)
        if (
            "StatusReason" in chset_data
            and "The submitted information didn't contain changes" in chset_data["StatusReason"]
        ):
            failed_for_real = False
        else:
            failed_for_real = True
        if "StatusReason" in chset_data:
            if failed_for_real:
                log_str = "\033[31;1mFAILED: " + chset_data["StatusReason"] + "\033[m"
            else:
                log_str = chset_data["StatusReason"]
            log(log_str)
        if failed_for_real:
            raise Exception("Creating changeset failed")
    else:
        chset_data["CreationTime"] = time.strftime(
            "%a, %d %b %Y %H:%M:%S +0000", chset_data["CreationTime"].timetuple()
        )
        log("\033[32;1m*** Changeset ***:\033[m")
        log_data(chset_data)
        if not dry_run:
            clf.execute_change_set(ChangeSetName=chset_id, DisableRollback=disable_rollback)
        else:
            clf.delete_change_set(ChangeSetName=chset_id)
    return


def create_stack(stack_name, template, params, session=None, tags=None, disable_rollback=False):
    params = get_template_arguments(stack_name, template, params)
    if tags:
        params["Tags"] = tags
    params["DisableRollback"] = disable_rollback
    cloudformation(session=session).create_stack(**params)
    return


def get_stack_operation(stack_name, session=None):
    stack_oper = "create_stack"
    try:
        stack_data = cloudformation(session=session).describe_stacks(StackName=stack_name)
        # Dump original status, for the record
        status = stack_data["Stacks"][0]["StackStatus"]
        log("Status: \033[32;1m" + status + "\033[m")
        stack_oper = "update_stack"
    except ClientError as err:
        if err.response["Error"]["Code"] == "ValidationError" and err.response["Error"]["Message"].endswith(
            "does not exist"
        ):
            log("Status: \033[32;1mNEW_STACK\033[m")
        else:
            raise
    return globals()[stack_oper]


def get_end_status(stack_name, session=None):
    logs = CloudWatchLogsThread(log_group_name=stack_name)
    logs.start()
    cf_events = CloudFormationEvents(log_group_name=stack_name)
    cf_events.start()
    log("Waiting for stack operation to complete:")
    status = "_IN_PROGRESS"
    while True:
        stack_info = cloudformation(session=session).describe_stacks(StackName=stack_name)
        status = stack_info["Stacks"][0]["StackStatus"]
        if "ROLLBACK" in status:
            color = "\033[31;1m"
        else:
            color = "\033[32;1m"
        log(color + "Status: " + status + "\033[m")
        if not status.endswith("_IN_PROGRESS"):
            logs.stop()
            cf_events.stop()
            break
        time.sleep(5)
    return status


def create_or_update_stack(stack_name, json_small, params_doc, session=None, tags=None, disable_rollback=False):
    stack_func = get_stack_operation(stack_name, session=session)
    stack_func(stack_name, json_small, params_doc, session=session, tags=tags, disable_rollback=disable_rollback)
    return get_end_status(stack_name, session=session)


def get_template_arguments(stack_name, template, params, session=None):
    params = {
        "StackName": stack_name,
        "Parameters": params,
        "Capabilities": ["CAPABILITY_IAM", "CAPABILITY_NAMED_IAM"],
    }
    if "CF_BUCKET" in os.environ and os.environ["CF_BUCKET"]:
        bucket = os.environ["CF_BUCKET"]
        template_hash = hashlib.md5()
        template_hash.update(template.encode("utf-8"))
        template_hash.update(aws_infra_util.json_save_small(params).encode("utf-8"))
        key = stack_name + "-" + template_hash.hexdigest()
        s3(session=session).put_object(Body=template, Bucket=bucket, Key=key)
        params["TemplateURL"] = "https://s3.amazonaws.com/" + bucket + "/" + key
    else:
        params["TemplateBody"] = template
    return params


def delete(stack_name, regn, session=None):
    os.environ["AWS_DEFAULT_REGION"] = regn
    log("**** Deleting stack '" + stack_name + "'")
    clf = cloudformation(session=session)
    cf_events = CloudFormationEvents(log_group_name=stack_name)
    cf_events.start()
    clf.delete_stack(StackName=stack_name)
    while True:
        try:
            stack_info = clf.describe_stacks(StackName=stack_name)
            status = stack_info["Stacks"][0]["StackStatus"]
            if not status.endswith("_IN_PROGRESS") and not status.endswith("_COMPLETE"):
                raise Exception("Delete stack failed: end state " + status)
            log("Status: \033[32;1m" + status + "\033[m")
            time.sleep(5)
        except ClientError as err:
            cf_events.stop()
            if err.response["Error"]["Code"] == "ValidationError" and err.response["Error"]["Message"].endswith(
                "does not exist"
            ):
                log("Status: \033[32;1mDELETE_COMPLETE\033[m")
                break
            else:
                raise


def resolve_ami(template_doc, session=None):
    ami_id = ""
    ami_name = ""
    ami_created = ""
    if "AMI_ID" in os.environ and os.environ["AMI_ID"]:
        ami_id = os.environ["AMI_ID"]

    if (
        not ami_id
        and "Parameters" in template_doc
        and "paramAmi" in template_doc["Parameters"]
        and "IMAGE_JOB" in os.environ
    ):
        image_job = re.sub(r"\W", "_", os.environ["IMAGE_JOB"].lower())
        log("Looking for ami with name prefix " + image_job)
        sorted_images = get_images(image_job)
        if sorted_images:
            image = sorted_images[0]
            ami_id = image["ImageId"]
            ami_name = image["Name"]
            ami_created = image["CreationDate"]
    elif ami_id and "Parameters" in template_doc and "paramAmi" in template_doc["Parameters"]:
        log("Looking for ami metadata with id " + ami_id)
        ami_meta = ec2(session=session).describe_images(ImageIds=[ami_id])
        log("Result: " + aws_infra_util.json_save(ami_meta))
        image = ami_meta["Images"][0]
        ami_name = image["Name"]
        ami_created = image["CreationDate"]
    return ami_id, ami_name, ami_created


def deploy(stack_name, yaml_template, regn, dry_run=False, session=None, disable_rollback=False):
    os.environ["AWS_DEFAULT_REGION"] = regn
    os.environ["REGION"] = regn
    global REDIRECTED
    if not REDIRECTED:
        # Disable buffering, from http://stackoverflow.com/questions/107705/disable-output-buffering
        sys.stdout = Unbuffered(sys.__stdout__)
        REDIRECTED = True
    template_doc = aws_infra_util.yaml_to_dict(yaml_template)
    ami_id, ami_name, ami_created = resolve_ami(template_doc, session=session)

    log(
        "**** Deploying stack '"
        + stack_name
        + "' with template '"
        + yaml_template
        + "' and ami_id '"
        + str(ami_id)
        + "'"
    )

    if "Parameters" not in template_doc:
        template_doc["Parameters"] = {}

    template_parameters = template_doc["Parameters"]
    tags = []
    if "Tags" in template_doc:
        tags = template_doc["Tags"]
        del template_doc["Tags"]
    if ami_id:
        with open("ami.properties", "w") as ami_props:
            ami_props.write("AMI_ID=" + ami_id + "\nNAME=" + ami_name + "\n")
        os.environ["paramAmi"] = ami_id
        os.environ["paramAmiName"] = ami_name
        os.environ["paramAmiCreated"] = ami_created
        if "paramAmiName" not in template_parameters:
            template_parameters["paramAmiName"] = collections.OrderedDict(
                [("Description", "AMI Name"), ("Type", "String"), ("Default", "")]
            )
        if "paramAmiCreated" not in template_parameters:
            template_parameters["paramAmiCreated"] = collections.OrderedDict(
                [
                    ("Description", "AMI Creation Date"),
                    ("Type", "String"),
                    ("Default", ""),
                ]
            )
    else:
        with open("ami.properties", "w") as ami_props:
            ami_props.write("AMI_ID=\nNAME=\n")
    json_small = aws_infra_util.json_save_small(template_doc)

    log("**** Final template ****")
    log_data(template_doc, output_format="json")

    # Create/update stack
    params_doc = []
    for key in list(template_parameters.keys()):
        if key in os.environ:
            val = os.environ[key]
            log("Parameter " + key + ": using \033[32;1mCUSTOM value " + val + "\033[m")
            params_doc.append({"ParameterKey": key, "ParameterValue": val})
        else:
            val = template_parameters[key]["Default"]
            log("Parameter " + key + ": using default value " + str(val))

    if not dry_run:
        status = create_or_update_stack(
            stack_name, json_small, params_doc, session=session, tags=tags, disable_rollback=disable_rollback
        )
        if not (status == "CREATE_COMPLETE" or status == "UPDATE_COMPLETE"):
            sys.exit("Stack operation failed: end state " + status)
    elif get_stack_operation(stack_name).__name__ == "update_stack":
        update_stack(stack_name, json_small, params_doc, dry_run=True, session=session, tags=tags)
    log("Done!")


class Unbuffered:
    def __init__(self, stream):
        self.stream = stream
        self.SYS_ENCODING = locale.getpreferredencoding(False)
        if self.SYS_ENCODING == "ascii" or self.SYS_ENCODING == "US-ASCII":
            self.SYS_ENCODING = "ISO8859-15"

    def write(self, data):
        if isinstance(data, str):
            self.stream.write(data)
        elif isinstance(data, bytes):
            self.stream.write(data.decode(self.SYS_ENCODING))
        else:
            self.stream.write(str(data))
        self.stream.flush()

    def writelines(self, datas):
        self.stream.writelines(datas)
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)
